:- use_module(library(lists)).

sat0(BoolExpr) :-
	term_variables(BoolExpr, Vs),
	maplist(bool, Vs),
	BoolExpr =:= 1.

bool(0).
bool(1).

sat(Expr) :-
	sat_to_constraints(Expr, Cs, 1),
	solve(Cs).

:- op(620, yfx, <=>).
:- op(610, yfx, =>).

defined_as((A => B), \A \/ B).
defined_as(A \/ B,   \(\A /\ \B)).
defined_as( A <=> B, (A => B)/\ (B => A)).

sat_to_constraints(Expr, Cs, Val) :-
	simple(Expr), !, Cs = [], Val = Expr.
sat_to_constraints(E1 <=> E2, Cs, Val) :-
	Val == 1, !, sat_to_constraints(E1, Cs1, V),
	sat_to_constraints(E2, Cs2, V),
	append(Cs1, Cs2, Cs).
sat_to_constraints(Expr0, Cs, Val) :-
	defined_as(Expr0, Expr), !,
	sat_to_constraints(Expr, Cs, Val).
sat_to_constraints(Expr, Cs, Val) :-
	Expr =.. [F|ExprArgs],
	maplist(sat_to_constraints, ExprArgs, ArgsCs, Args),
	ExprV =.. [F|Args], 
	append([[constraint(Val, ExprV)]|ArgsCs], Cs).

% solve(Cs0): repeatedly applies simplify until no change occurs.
% Then it sets the first variable to O or 1 and recurses.
solve([]) :- !.
solve(Cs0) :-
	simplify(Cs0, Cs), !,
	(   Cs == Cs0 
	->  term_variables(Cs0, [Var|_]),
	    bool(Var), solve(Cs0)
	;   solve(Cs)
	).

% simplify(Cs0, Cs): Cs is obtained by possibly reducing each element of Cs0.
% Fails if a reduction detects failure.
simplify([], Cs) :- !, Cs = [].
simplify([C0|Cs0], Cs) :-
	reducible(C0, Test), !,
	Test, simplify(Cs0, Cs).
simplify([C|Cs0], [C|Cs]) :-
	simplify(Cs0, Cs).

% reducible(Constraint, Test): succeeds if the truth value of Constraint
% can be obtained by running Test.
reducible(constraint(X, A /\ B), Test) :-
	(   A == 0 -> Test = (X = 0)
	;   A == 1 -> Test = (X = B)
	;   B == 0 -> Test = (X = 0)
	;   B == 1 -> Test = (X = A)
	;   X == 1 -> Test = (A = 1, B = 1)
	;   B == A -> Test = (X = A)
	).
reducible(constraint(X, \ A), Test) :-
	(   nonvar(X) -> NotX is 1-X, Test = (A = NotX)
	;   nonvar(A) -> NotA is 1-A, Test = (X = NotA)
	;   X == A    -> Test = fail
	).

end_of_file.

% Az albbi nyomkvetsben a numerikus vltozneveket (remlhetleg)
% rtelmes nevekre rtuk t

| ?- trace, sat(\ (\A /\ \B) /\ \A).
% The debugger will first creep -- showing everything (trace)
        1      1 Call: sat(\ (\A/\ \B)/\ \A) ? 
        2      2 Call: sat_to_constraints(\ (\A/\ \B)/\ \A,_3035,1) ? s
        2      2 Exit: sat_to_constraints(\ (\A/\ \B)/\ \A,[constraint(1,AorB/\NA),constraint(AorB,\AnorB),constraint(AnorB,NA2/\NB),constraint(NA2,\A),constraint(NB,\B),constraint(NA,\A)],1) ? 
        3      2 Call: solve([constraint(1,AorB/\NA),constraint(AorB,\AnorB),constraint(AnorB,NA2/\NB),constraint(NA2,\A),constraint(NB,\B),constraint(NA,\A)]) ? 
        4      3 Call: simplify([constraint(1,AorB/\NA),constraint(AorB,\AnorB),constraint(AnorB,NA2/\NB),constraint(NA2,\A),constraint(NB,\B),constraint(NA,\A)],_5457) ? s
        4      3 Exit: simplify([constraint(1,1/\1),constraint(1,\0),constraint(0,NA2/\NB),constraint(NA2,\0),constraint(NB,\B),constraint(1,\0)],[constraint(0,NA2/\NB),constraint(NA2,\0),constraint(NB,\B)]) ? 
        5      3 Call: solve([constraint(0,NA2/\NB),constraint(NA2,\0),constraint(NB,\B)]) ? 
        6      4 Call: simplify([constraint(0,NA2/\NB),constraint(NA2,\0),constraint(NB,\B)],_7481) ? s
        6      4 Exit: simplify([constraint(0,1/\NB),constraint(1,\0),constraint(NB,\B)],[constraint(0,1/\NB),constraint(NB,\B)]) ? 
        7      4 Call: solve([constraint(0,1/\NB),constraint(NB,\B)]) ? 
        8      5 Call: simplify([constraint(0,1/\NB),constraint(NB,\B)],_9451) ? s
        8      5 Exit: simplify([constraint(0,1/\0),constraint(0,\1)],[]) ? 
        9      5 Call: solve([]) ? 
        9      5 Exit: solve([]) ? 
        7      4 Exit: solve([constraint(0,1/\0),constraint(0,\1)]) ? 
        5      3 Exit: solve([constraint(0,1/\0),constraint(1,\0),constraint(0,\1)]) ? 
        3      2 Exit: solve([constraint(1,1/\1),constraint(1,\0),constraint(0,1/\0),constraint(1,\0),constraint(0,\1),constraint(1,\0)]) ? 
        1      1 Exit: sat(\ (\0/\ \1)/\ \0) ? 
A = 0,
B = 1 ? ;
no
| ?- sat(A <=> \A /\ \B).
        1      1 Call: sat(A<=> \A/\ \B) ? 
        2      2 Call: sat_to_constraints(A<=> \A/\ \B,_2851,1) ? s
        2      2 Exit: sat_to_constraints(A<=> \A/\ \B,[constraint(A,NA/\NB),constraint(NA,\A),constraint(NB,\B)],1) ? 
        3      2 Call: solve([constraint(A,NA/\NB),constraint(NA,\A),constraint(NB,\B)]) ? 
        4      3 Call: simplify([constraint(A,NA/\NB),constraint(NA,\A),constraint(NB,\B)],_5051) ? s
        4      3 Exit: simplify([constraint(A,NA/\NB),constraint(NA,\A),constraint(NB,\B)],[constraint(A,NA/\NB),constraint(NA,\A),constraint(NB,\B)]) ? 
        5      3 Call: term_variables([constraint(A,NA/\NB),constraint(NA,\A),constraint(NB,\B)],[_6363|_6365]) ? s
        5      3 Exit: term_variables([constraint(A,NA/\NB),constraint(NA,\A),constraint(NB,\B)],[A,NA,NB,B]) ? 
        6      3 Call: bool(A) ? s
?       6      3 Exit: bool(0) ? 
        7      3 Call: solve([constraint(0,NA/\NB),constraint(NA,\0),constraint(NB,\B)]) ? 
        8      4 Call: simplify([constraint(0,NA/\NB),constraint(NA,\0),constraint(NB,\B)],_9653) ? s
        8      4 Exit: simplify([constraint(0,1/\NB),constraint(1,\0),constraint(NB,\B)],[constraint(0,1/\NB),constraint(NB,\B)]) ? 
        9      4 Call: solve([constraint(0,1/\NB),constraint(NB,\B)]) ? s
        9      4 Exit: solve([constraint(0,1/\0),constraint(0,\1)]) ? 
        7      3 Exit: solve([constraint(0,1/\0),constraint(1,\0),constraint(0,\1)]) ? 
?       3      2 Exit: solve([constraint(0,1/\0),constraint(1,\0),constraint(0,\1)]) ? 
?       1      1 Exit: sat(0<=> \0/\ \1) ? 
A = 0,
B = 1 ? ;
        1      1 Redo: sat(0<=> \0/\ \1) ? 
        3      2 Redo: solve([constraint(0,1/\0),constraint(1,\0),constraint(0,\1)]) ? 
        6      3 Redo: bool(0) ? 
        6      3 Exit: bool(1) ? 
       10      3 Call: solve([constraint(1,NA/\NB),constraint(NA,\1),constraint(NB,\B)]) ? 
       11      4 Call: simplify([constraint(1,NA/\NB),constraint(NA,\1),constraint(NB,\B)],_9647) ? s
       11      4 Fail: simplify([constraint(1,NA/\NB),constraint(NA,\1),constraint(NB,\B)],_9647) ? 
       10      3 Fail: solve([constraint(1,NA/\NB),constraint(NA,\1),constraint(NB,\B)]) ? 
        3      2 Fail: solve([constraint(A,NA/\NB),constraint(NA,\A),constraint(NB,\B)]) ? 
        1      1 Fail: sat(A<=> \A/\ \B) ? 
no


